/*
Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

*/
#include "Cfn.hpp"

std::vector<std::map<double,double> > Cfn(lattice_graph& LG, uint na,uint nb, uint nc, std::string bc, std::string lattice_pars)
{
  edge_iterator ea,eb,ebegin,eend; 
  site_iterator s,t,last_site,first_site;
  int i;
  double p;
  bool inserted;

  // edges and distances list are made here
  // copy from build_lattice, with cut_off switched off.

  double distance;
  euclid_vector index_difference,translate; 
  lattice_vectors lv=get_lattice_vectors(lattice_pars);
  //connect all lattice sites to all other lattice sites, and 
  //calculate their mutual disances, across lattice boundaries if
  //periodic boundary conditions apply.

  std::vector<std::map<double,double> > Cfn(2);
  std::map<double,double> cnt;

  Edge edge;
  for ( tie(s,last_site)=vertices(LG); s!=(last_site-1); s++) {
    for (t=s+1; t!=last_site; t++) {
      translate=zero; 
      index_difference = LG[*t].cell_index + LG[*t].fract_coords - 
	LG[*s].cell_index - LG[*s].fract_coords;
      if ( (fabs(index_difference[A]) > (1.0*na)/2.0) & (bc.find("a") < 10 )) 
	translate[A]=index_difference[A]*na/fabs(index_difference[A]); 
      if ( (fabs(index_difference[B]) > (1.0*nb)/2.0) & (bc.find("b") < 10 )) 
	translate[B]=index_difference[B]*nb/fabs(index_difference[B]);
      if ( (fabs(index_difference[C]) > (1.0*nc)/2.0) & (bc.find("c") < 10 )) 
	translate[C]=index_difference[C]*nc/fabs(index_difference[C]);

      distance = fixed(ublas::norm_2((LG[*t].x - LG[*s].x) - 
	(translate[A]*lv[A] + translate[B]*lv[B] + translate[C]*lv[C])),3);

      //The correlation function should be calculated here of
      //course. That saves a lot of memory! (as long as there is no
      //double counting.)

      p = ublas::inner_prod(LG[*s].spin,LG[*t].spin); 
      Cfn[0][distance]+=p;
      Cfn[1][distance]+=p*p;
      cnt[distance]++;
      /*
	if (ublas::norm_2(distance) < cutoff) { 
	  tie(edge,inserted)=add_edge(*s,*t,LG);
      
	  LG[edge].ru=distance; 
	  LG[edge].r_abs=fixed(ublas::norm_2(distance),3);
	  LG[edge].ru/=LG[edge].r_abs;
	}
      */
    }
  }  


  //std::cout<<"ok\nNow calculate the correlation function.\n"<<std::endl;
  // Now this is all that is left to calculate the pairwise
  // correlation function.

  /*
  tie(ebegin,eend)=edges(LG);
  for (ea=ebegin; ea!=eend; ea++) {
    Cfn[LG[*ea].r_abs]+=ublas::inner_prod(LG[source(*ea,LG)].spin,
					  LG[target(*ea,LG)].spin);
    cnt[LG[*ea].r_abs]++;
    }
  */
  //std::cout<<"Result still needs to be normalised..."<<std::endl;
  
  std::map<double,double>::iterator Cfn_it;
  std::map<double,double>::iterator cnt_it;
  cnt_it = cnt.begin();

  //std::cout<<Cfn.size()<<" "<<cnt.size()<<"
  //"<<(*Cfn_it).first<<"..ok\n";
  for (i=0; i<2; i++) {
    Cfn_it = Cfn[i].begin();  
    for (;Cfn_it!=Cfn[i].end();Cfn_it++) {
      //std::cout<<(*Cfn_it).first<<(*Cfn_it).second<<std::endl;
      (*Cfn_it).second/=(*cnt_it).second;
      cnt_it++;
    }
    Cfn[i][0] = 1;
  }

  //std::cout<<"ok, that should be it.\n"<<std::endl;
  //std::sort(Cfn.begin(),Cfn.end());

  return Cfn;
}

lattice_graph Chirality(po::variables_map vm, lattice_graph& LG)
{
  //Calculate the cross correlation function of the chirality
  edge_iterator ea, eb, ebegin, eend; 
  site_iterator s,last_site,first_site;
  graph_traits<lattice_graph>::adjacency_iterator t,firstneighbour,lastneighbour;

  euclid_vector S1, S2, S3;
  lattice_graph XG;
  boost::graph_traits<lattice_graph>::vertex_descriptor X;

  std::vector<int> S1site=vm["S1.site"].as<std::vector<int> >();
  std::vector<int> S2site=vm["S2.site"].as<std::vector<int> >();
  std::vector<int> S3site=vm["S3.site"].as<std::vector<int> >();

  std::vector<std::string> S1cell(vm["S1.cell"].as<std::vector<std::string> >() );
  std::vector<std::string> S2cell(vm["S2.cell"].as<std::vector<std::string> >() );
  std::vector<std::string> S3cell(vm["S3.cell"].as<std::vector<std::string> >() );
  euclid_vector supercell = parse_vector(3,vm["supercell"].as<std::string> () );

  if ((S1site.size()!=S2site.size())| (S1site.size()!=S3site.size()) |
      (S1site.size()!=S2cell.size())| (S1site.size()!=S3cell.size()))
    po::invalid_option_value("S1sites, to_sites 2,3 and to_cells 2,3 must all have same number of entries"); 

  tie(first_site,last_site)=vertices(LG);
  tie(ebegin,eend)=edges(LG);
  //search for S1sites:
  for (s=first_site;s!=last_site;s++) //loop over all vertices in lattice
    for (int i=0; i<S1site.size();i++){ //loop over all S1sites specified in config{
      if ((LG[*s].site_index==S1site[i]) & 
	  ((LG[*s].cell_index-parse_vector(3,S1cell[i]))%supercell==zero)){
	//find targets for this S1 site, and get their spins. 

	S1=LG[*s].spin;
	tie(firstneighbour,lastneighbour)=adjacent_vertices(*s,LG);
	//std::cout<<"Now looping over neighbours for S1site"<<int(*s)<<std::endl;
	for (t=firstneighbour;t!=lastneighbour;t++) {
	  euclid_vector translation=LG[*t].cell_index-LG[*s].cell_index;
	  if ((LG[*t].site_index==S2site[i])&&(translation==(parse_vector(3,S2cell[i])) ))
	    S2=LG[*t].spin;
	  else if ((LG[*t].site_index==S3site[i])&&(translation==(parse_vector(3,S3cell[i])) ))
	    S3=LG[*t].spin;
	}
	// now make add a vertex to the new graph, with the position of the S1site in
	// the original graph, and as spin the chirality calculated from S1, S2 and S3.
	X = add_vertex(XG);
	//std::cout<<"Added vertex in new graph corresponding to this S1site.\n";
	XG[X].spin = 2/(3*sqrt(3))*(outer_product(S1,S2) +
				    outer_product(S2,S3) + outer_product(S3,S1));
	XG[X].cell_index=LG[*s].cell_index; 
	XG[X].site_index=i; 
	XG[X].x=LG[*s].x;
	XG[X].fract_coords=LG[*s].fract_coords;
	XG[X].mu_eff=1;
	XG[X].label="none";
      }
    }
  //calculate the pairwise correlation function of this new graph. 

  return XG;
}

lattice_graph Nematicity(po::variables_map vm, lattice_graph& LG)
{
  //Calculate the spin nematic correlation function of the chirality
  edge_iterator ea, eb, ebegin, eend; 
  site_iterator s,last_site,first_site;
  graph_traits<lattice_graph>::adjacency_iterator t,firstneighbour,lastneighbour;

  euclid_vector S1, S2;
  lattice_graph XG;
  boost::graph_traits<lattice_graph>::vertex_descriptor X;

  std::vector<int> S1site=vm["S1.site"].as<std::vector<int> >();
  std::vector<int> S2site=vm["S2.site"].as<std::vector<int> >();
  std::vector<std::string> S1cell(vm["S1.cell"].as<std::vector<std::string> >() );
  std::vector<std::string> S2cell(vm["S2.cell"].as<std::vector<std::string> >() );
  euclid_vector supercell = parse_vector(3,vm["supercell"].as<std::string> () );

  if ( (S1site.size()!=S2site.size()) | (S1site.size()!=S2cell.size()) )
    po::invalid_option_value("S1sites, to_sites 2,3 and to_cells 2,3 must all have same number of entries"); 
  tie(first_site,last_site)=vertices(LG);
  tie(ebegin,eend)=edges(LG);
  //search for S1sites:
  for (s=first_site;s!=last_site;s++) //loop over all vertices in lattice
    for (int i=0; i<S1site.size();i++) {//loop over all S1sites specified in config{
      if ((LG[*s].site_index==S1site[i])&
	  ((LG[*s].cell_index-parse_vector(3,S1cell[i]))%supercell==zero)){
	//find targets for this S1site, and get their spins. 
	S1=LG[*s].spin;
	tie(firstneighbour,lastneighbour)=adjacent_vertices(*s,LG);
	//std::cout<<"Now looping over neighbours for S1site"<<int(*s)<<std::endl;
	for (t=firstneighbour;t!=lastneighbour;t++) {
	  euclid_vector translation=LG[*t].cell_index-LG[*s].cell_index;
	  if ((LG[*t].site_index==S2site[i])&&(translation==parse_vector(3,S2cell[i]) ))
	    S2=LG[*t].spin;
	}
	// now make add a vertex to the new graph, with the position of the S1site in
	// the original graph, and as spin the chirality calculated from S1, S2 and S3.
	X = add_vertex(XG);
	//std::cout<<"Added vertex in new graph corresponding to this S1site.\n";
	XG[X].spin = outer_product(S1,S2);

	XG[X].cell_index=LG[*s].cell_index; 
	XG[X].site_index=i; 
	XG[X].x=LG[*s].x;
	XG[X].fract_coords=LG[*s].fract_coords;
	XG[X].mu_eff=1;
      }
    }
  return XG;
}
