#ifndef FLIPPERS
#define FLIPPERS
/*  
Spinner v.2

A Monte Carlo program for the simulation of 'classical'
magnetic systems.

Written by Mark de Vries, markadevries@gmail.com

Copyright (c), Aug 2005-Jan 2007, Dec 2012, The University Court of the University of
Edinburgh, Old College, Edinburgh, UK  

Copyright (c), Aug 2005-Oct 2006, Council for the Central Laboratory of the 
Research  Councils, Rutherford Appleton Laboratory, Chilton, Didcot, Oxfordshire, UK

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation; either version 2 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include <iostream>
#include <stdio.h>
#include <stdlib.h>
#include <boost/random/mersenne_twister.hpp>
#include <boost/random/lagged_fibonacci.hpp> 
#include <boost/random/variate_generator.hpp>
#include <boost/random/uniform_on_sphere.hpp>
#include <boost/random/uniform_int.hpp>
#include <boost/random/uniform_real.hpp>

struct flipper_base
{
public:
  flipper_base(){;};
  void operator()(site_iterator& site, lattice_graph& LG,euclid_vector& R,double phim)
  {;
  };
};

struct Potts
{
public:
  Potts(){;};
  void operator()(site_iterator& site, lattice_graph& LG,euclid_vector& R,double phim)
  {
    if (int(LG[*site].spin[0])==1) 
      LG[*site].spin[0]=-0.5+0.5*R[0];
    else if (int(LG[*site].spin[0])==0) 
      LG[*site].spin[0]=R[0];
    else if (int(LG[*site].spin[0])==-1) 
      LG[*site].spin[0]=0.5+0.5*R[0];
    //S[0]=int(S[0])+1.5+0.5*R[0];
    //if (int(S[0])>1) 
    //  S[0]=int(S[0]-3);
  };
};

struct Ising
{ 
public:
  Ising(){std::cout<<"\nSIMULATION TYPE: Ising\n";};
  void operator()(site_iterator& site, lattice_graph& LG,euclid_vector& R,double phi_max)
  {LG[*site].spin=-LG[*site].spin;};
};

struct XY
{
public:
  XY():S(2){std::cout<<"\nSIMULATION TYPE: XY\n";};
  void operator()(site_iterator& site, lattice_graph& LG,euclid_vector& R,double phi_max)
  {
    S = LG[*site].spin + phi_max*R; //Note:alternative spin-flip method. 
    LG[*site].spin=S/ublas::norm_2(S);
  };
private:
  euclid_vector S;
};

struct Heisenberg
{
public:
  Heisenberg();
  void operator()(site_iterator& site, lattice_graph& LG,euclid_vector R,double phi_max)
  {
    S = LG[*site].spin;
    phi=phi_max*uni_pm1();
    LG[*site].spin = cos(phi) * S + (ublas::inner_prod(R,S)*(1-cos(phi))) * R + 
      sin(phi) * outer_product(R,S); 
  };
private:
  double phi;
  lagged_fibonacci607 fib_gen;  
  uniform_real<long double> uni_pm1_dist;
  euclid_vector S;
  variate_generator<lagged_fibonacci607&, uniform_real<long double> > uni_pm1;
};


Heisenberg::Heisenberg() :
  uni_pm1_dist(-1,1), uni_pm1(fib_gen,uni_pm1_dist),S(3)
{
  fib_gen.seed(static_cast<uint>(time(0))+1000);
  std::cout<<"\nSIMULATION TYPE: Heisenberg\n";
}


#endif
